//
// File: gpssync.cc
// GPS sync program $Id: gpssync.cc,v 2.0 1999/03/30 12:43:39 lambert Exp $
// David Lambert 
// February 16, 1999 
//
// This program waits for and reads the NMEA RMC sentence from a GPS device.
// The time and date are extracted and converted into the form MMDDhhmmYY.ss
// and printed on stdout. This is the format that the program 'date'
// requires for setting the system time. 
// 
// $Log: gpssync.cc,v $
// Revision 2.0  1999/03/30 12:43:39  lambert
// Release 2.0
//
// Revision 1.11  1999/03/29 22:59:04  lambert
// Corrected GPS time printout
//
// Revision 1.10  1999/03/29 22:12:23  lambert
// Whoops! Checked it in too soon (before testing).
//
// Revision 1.9  1999/03/29 22:09:12  lambert
// Added code to handle years greater than 99 correctly.
// i.e. Y2k issues
//
// Revision 1.8  1999/03/28 14:35:50  lambert
// Removed logging to stderr except in DEBUG mode.
//
// Revision 1.7  1999/03/28 04:16:26  lambert
// Added printout of time difference between GPS
// and CPU time.
//
// Revision 1.6  1999/03/27 01:46:25  lambert
// *** empty log message ***
//
// Revision 1.5  1999/03/27 01:14:30  lambert
// Added syslog messages in anticipation of this being used as a
// kernel cron job.
//
// Revision 1.4  1999/03/25 23:53:10  lambert
// Added error exits.
//
// Revision 1.3  1999/03/25 23:18:15  lambert
// Changed the init function to be more in line with the
// reccommendations of the Serial Programming HOWTO
//
//

#include <stdio.h> 
#include <string.h> 
#include <sys/stat.h> 
#include <sys/types.h> 
#include <fcntl.h> 
#include <termios.h> 
#include <unistd.h> 
#include <sys/signal.h> 
#include <time.h> 
#include <signal.h>
#include <syslog.h>

static void init(char *dev); 
static void errorexit(int err);
static void timeout(int err);

static int fd; 
static FILE *gps; 
static struct termios oldtio,newtio;




int main(int argc, char **argv) 
{ 
  static struct sigaction action = { // For timeout
    timeout,
    {{0}},
    0,
    NULL
  };				// Alarm signal handler
  struct tm gpstime;		// For computing time adjustment
  time_t cputime;
  int hour,min,sec,day,mon,year; 
  char sentence[128];		// Holder for NMEA sentence from GPS
  openlog("gpssync",		// Set up syslog
	  LOG_PERROR |		// Also log to stderr
	  LOG_PID,LOG_USER);
  if (argc < 2) { 
    syslog(LOG_CRIT,"Need port as first parameter"); 
    exit(-1); 
  }
  sigaction(SIGALRM,&action,NULL); // Set up the alarm signal handler
  init(argv[1]);		// Initialize the com port 
  do {
    alarm(5);			// Set up a five second timer
    if (fgets(sentence,sizeof(sentence),gps) == NULL) { // Get an NMEA sentence 
      syslog(LOG_CRIT,"Error getting sentence: %m"); 
      errorexit(-1); 
    } 
#ifdef DEBUG
    printf("Sentence = %s\n",sentence); // Debug 
#endif
  } while (strncmp("$GPRMC",sentence,6) != 0); // Until have the RMC sentence 
  unsigned int i=1;		// Loop count (skipping $) 
  unsigned int csumacc=0;	// Checksum accumulator 
  unsigned int csum=0xff;	// Checksum from message 
  do { 
    csumacc^=sentence[i];	// Compute checksum
  } while ((i < sizeof(sentence)) && // Check that we are not running off the end. 
	   (sentence[++i] != '*')); // Up to the marker 
  if (sentence[i] == '*') {	// We found the marker 
    char status;
    sscanf(&sentence[i+1],"%2x",&csum); // Read the given checksum 
    if (csum == csumacc) {	// We are ok. let's start the analysis
      i=0;
      int parcount=0;
      char *sp=sentence;	// Point to start of sentence
      do {			// get the parameters
	if (*sp++ == ',') {	// We have a parameter
	  parcount++;		// Count it
	  if (parcount == 1) {	// This is the time
	    sscanf(sp," %2d%2d%2d",&hour,&min,&sec);
	  }
	  else if (parcount == 2) { // This is the gps status
	    sscanf(sp,"%c",&status);
	  }
	  else if (parcount == 9) { // This is the date
	    sscanf(sp,"%2d%2d%2d",&day,&mon,&year);
	  }
	}
      } while (*sp != '*');
      if (status != 'A') {
	syslog(LOG_CRIT,"GPS has no navigation fix - time is unreliable.");
	errorexit(-1);
      }
      printf("%02d%02d%02d%02d%02d.%02d\n",mon,day,hour,min,year,sec); // This is for 'date' program
      time(&cputime);		// Get the current cpu time
      bzero(&gpstime,sizeof(gpstime)); // Now for the diagnostics
      gpstime.tm_sec=sec;	// Populate tm struct with gps info.
      gpstime.tm_min=min;
      gpstime.tm_hour=hour;
      gpstime.tm_mday=day;
      gpstime.tm_mon=mon-1;
      gpstime.tm_year=year;
      if (gpstime.tm_year < 70) { // Y2k issues
	gpstime.tm_year += 100;	// Into the next millenium! (1900 + 100=2000)
      }
      time_t gpstime_t = mktime(&gpstime); // Get gpstime as time_t
      syslog(LOG_INFO,
	     "Clock adjustment %d seconds, UTC time from GPS is: %s",
	     gpstime_t - mktime(gmtime(&cputime)), // Compute difference
	     ctime(&gpstime_t)); // Printed UTC time
    } 
    else {			// Bad checksum 
      syslog(LOG_CRIT,"Bad Checksum, Read %2x, Computed %2x",csum,csumacc);
      errorexit(-1);
    } 
  }
  else {			// No marker 
    syslog(LOG_CRIT,"No end of message marker found"); 
  }
  tcsetattr(fd,TCSANOW,&oldtio); // Set line back to saved state
  return 0;			// Success
}
 
//
// Initialize the communication line to the GPS as specified in 'dev. 
// Parameters are as specified in NMEA-0182
//

static void init(char *dev) 
{
  //    Open modem device for reading and writing and not as controlling tty
  //    because we don't want to get killed if linenoise sends CTRL-C.

   fd = open(dev, O_RDONLY | O_NOCTTY );
   if (fd <0) {
     syslog(LOG_CRIT,"Open error on %s: %m",dev); 
     errorexit(-1);
   }
   tcgetattr(fd,&oldtio); // save current serial port settings 
   bzero(&newtio, sizeof(newtio)); // clear struct for new port settings 


   //    B4800 Set baud rate to 4800
   //    CRTSCTS : output hardware flow control (only used if the cable has
   //    all necessary lines. See sect. 7 of Serial-HOWTO)
   //    CS8     : 8n1 (8bit,no parity,1 stopbit)
   //    CLOCAL  : local connection, no modem contol
   //    CREAD   : enable receiving characters

   newtio.c_cflag = B4800 | CRTSCTS | CS8 | CLOCAL | CREAD;


   //    IGNPAR  : ignore bytes with parity errors
   //    ICRNL   : map CR to NL (otherwise a CR input on the other computer
   //              will not terminate input)
   //    otherwise make device raw (no other input processing)

   newtio.c_iflag = IGNPAR | ICRNL;


   //   Raw output.

   newtio.c_oflag = 0;


   //    ICANON  : enable canonical input
   //    disable all echo functionality, and don't send signals to calling program

   newtio.c_lflag = ICANON;

   //    initialize all control characters
   //    default values can be found in /usr/include/termios.h, and are given
   //    in the comments, but we don't need them here

   newtio.c_cc[VINTR]    = 0;     // Ctrl-c 
   newtio.c_cc[VQUIT]    = 0;     // Ctrl-\ 
   newtio.c_cc[VERASE]   = 0;     // del 
   newtio.c_cc[VKILL]    = 0;     // @ 
   newtio.c_cc[VEOF]     = 4;     // Ctrl-d 
   newtio.c_cc[VTIME]    = 0;     // inter-character timer unused 
   newtio.c_cc[VMIN]     = 1;     // blocking read until 1 character arrives 
   newtio.c_cc[VSWTC]    = 0;     // '\0' 
   newtio.c_cc[VSTART]   = 0;     // Ctrl-q 
   newtio.c_cc[VSTOP]    = 0;     // Ctrl-s 
   newtio.c_cc[VSUSP]    = 0;     // Ctrl-z 
   newtio.c_cc[VEOL]     = 0;     // '\0' 
   newtio.c_cc[VREPRINT] = 0;     // Ctrl-r 
   newtio.c_cc[VDISCARD] = 0;     // Ctrl-u 
   newtio.c_cc[VWERASE]  = 0;     // Ctrl-w 
   newtio.c_cc[VLNEXT]   = 0;     // Ctrl-v 
   newtio.c_cc[VEOL2]    = 0;     // '\0' 


   //    now clean the modem line and activate the settings for the port

   tcflush(fd, TCIFLUSH);
   tcsetattr(fd,TCSANOW,&newtio);

   if ((gps=fdopen(fd,"r")) == NULL) { // Open stream 
     syslog(LOG_CRIT,"Error on fdopen: %m"); 
     errorexit(-1); 
   } 
} 

//
// Function to exit the program on error after setting the line back to its saved state
//

static void errorexit(int err)
{
  tcsetattr(fd,TCSANOW,&oldtio); // Set line back to saved state
  exit(err);
}


//
// Timeout handler
//
static void timeout(int err)
{
  syslog(LOG_CRIT,"GPS not responding - check connections");
  errorexit(err);
}
